# SPDX-FileCopyrightText: Emery Hemingway
#
# SPDX-License-Identifier: LicenseRef-Hippocratic-1.1

{ testPkgs, hostPkgs, lib }:

let
  testDriver = with hostPkgs;
    stdenv.mkDerivation {
      name = "nova-genode-test-driver";

      buildInputs = [ makeWrapper expect ];

      dontUnpack = true;

      preferLocalBuild = true;

      installPhase = ''
        install -Dm555 ${./nova-test-driver.exp} $out/bin/genode-test-driver
        wrapProgram $out/bin/genode-test-driver \
          --prefix PATH : "${lib.makeBinPath [ expect coreutils ]}"
      '';
    };

  runTests = driver:
    hostPkgs.stdenv.mkDerivation {
      name = "test-run-${driver.testName}";

      buildCommand = ''
        mkdir -p $out/nix-support

        ${driver}/bin/genode-test-driver | tee $out/log

        touch $out/nix-support
        echo "report testlog $out log" >> $out/nix-support/hydra-build-products
      '';
    };

  mkIso = import ./nova-iso.nix {
    inherit hostPkgs testPkgs;
    coreNovaObj = "${testPkgs.depot.base-nova}/lib/core-nova.o";
  };

  makeTest = { name ? "unamed", testScript, testConfig, bootModules
    , qemuMem ? 32, ... }@t:
    let
      iso = mkIso (with testPkgs;
        {
          inherit testConfig;
          config = ./driver-config.xml;
          init = "${genode.os}/bin/init";
          "ld.lib.so" = "${depot.base-nova}/lib/ld.lib.so";
          timer = "${genode.base-nova}/bin/nova_timer_drv";
        } // bootModules);

      baseSetup = ''
        ##
        # Wait for a specific output of a already running spawned proce
        #
        proc wait_for_output { wait_for_re timeout_value running_spawn_id } {
        	global output

        	if {$wait_for_re == "forever"} {
        		set timeout -1
        		interact {
        			\003 {
        				send_user "Expect: 'interact' received 'strg+c' and was cancelled\n";
        				exit
        			}
        			-i $running_spawn_id
        		}
        	} else {
        		set timeout $timeout_value
        	}

        	expect {
        		-i $running_spawn_id -re $wait_for_re { }
        		eof     { puts stderr "Error: Spawned process died unexpectedly"; exit -1 }
        		timeout { puts stderr "Error: Test execution timed out"; exit -1 }
        	}
        	set output $expect_out(buffer)
        }


        proc run_genode_until {{wait_for_re forever} {timeout_value 0} {running_spawn_id -1}} {
        	#
        	# If a running_spawn_id is specified, wait for the expected output
        	#
        	if {$running_spawn_id != -1} {
        		wait_for_output $wait_for_re $timeout_value $running_spawn_id
        		return
        	}

        	global env
        	set out $env(out)

        	set fd [open "$out/nix-support/hydra-build-products" w]
        	puts $fd "file iso ${iso}"
        	close $fd

        	global spawn_id
        	spawn ${hostPkgs.qemu_test}/bin/qemu-system-x86_64 -cdrom ${iso} -nographic \
        		-m size=${toString qemuMem}
        	wait_for_output $wait_for_re $timeout_value $spawn_id
        }
      '';

      driver = with hostPkgs;
        buildPackages.runCommand "genode-test-driver-${name}" {
          buildInputs = [ makeWrapper expect ];
          inherit baseSetup testConfig testScript;
          preferLocalBuild = true;
          testName = name;
        } ''
          mkdir -p $out/bin
          echo "$testConfig" > $out/test.config
          echo "$testScript" > $out/test-script
          echo "$baseSetup" > $out/base-setup
          ln -s ${testDriver}/bin/genode-test-driver $out/bin/
          wrapProgram $out/bin/genode-test-driver \
            --run "export testConfig=\"$testConfig\"" \
            --run "export testScript=\"\$(cat $out/test-script)\"" \
            --run "export baseSetup=\"\$(cat $out/base-setup)\"" \
        '';

      passMeta = drv:
        drv
        // lib.optionalAttrs (t ? meta) { meta = (drv.meta or { }) // t.meta; };

      test = passMeta (runTests driver);

    in test // { inherit driver test; };

in {
  callTest = path: args:
    makeTest (import path ({
      pkgs = testPkgs;
      inherit lib;
    } // args));
}
